/*
 * Copyright (c) 2002, 2003, 2004
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2002, 2003, 2004\n\
The Regents of the University of California.  All rights reserved.\n";
static const char rcsid[] =
    "@(#) $Id: acld.c,v 1.26 2004/08/09 20:27:23 leres Exp $ (LBL)";
#endif

/*
 * acld - manage acl's on a router
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "gnuc.h"
#ifdef HAVE_OS_PROTO_H
#include "os-proto.h"
#endif

#include "acld.h"
#include "cf.h"
#include "child.h"
#include "server.h"
#include "setsignal.h"
#include "util.h"

#define MAX(x, y) ((x) > (y) ? (x) : (y))

/* Globals */
int debug;			/* turn on debugging and don't fork */
int verbose;			/* not really used for anything! */
int reload;			/* kill child and reload config */
int foreground;
const char *prog;
const char *configfn = "/usr/local/etc/acld.conf";

struct s2v astate2str[] = {
	{ "notconnected",	ASTATE_NOTCONNECTED },
	{ "connected",		ASTATE_CONNECTED },
	{ "loggedin",		ASTATE_LOGGEDIN },
	{ "sentlogin",		ASTATE_SENTLOGIN },
	{ "sentlistacl",	ASTATE_SENTLISTACL },
	{ "readresponse",	ASTATE_READRESPONSE },
	{ "readerror",		ASTATE_READERROR },
	{ "readacl",		ASTATE_READACL },
	{ NULL,			0 }
};

struct s2v rstate2str[] = {
	{ "pending",		RSTATE_PENDING },
	{ "readcomment",	RSTATE_READCOMMENT },
	{ "readresponse",	RSTATE_READRESPONSE },
	{ "child",		RSTATE_CHILD },
	{ "done",		RSTATE_DONE },
	{ NULL,			0 }
};

/* Locals */
static struct state state;
static int backlog = 16;		/* listen() backlog */
static int sawterm;

static const char *pidfile = "/var/run/acld.pid";

/* Externs */
extern int optind;
extern int opterr;
extern char *optarg;

/* Forwards */
int expect(struct state *);
int main(int, char **);
void settimeout(struct state *);
RETSIGTYPE sigchild(int);
RETSIGTYPE sighup(int);
#ifdef notdef
RETSIGTYPE sigpipe(int);
#endif
RETSIGTYPE sigterm(int);
__dead void usage(void) __attribute__((volatile));

int
main(int argc, char **argv)
{
	register int n, c, op;
	register char *cp;
	register time_t t;
	register FILE *fp;
	register ssize_t cc;
	register struct req *rp;
	register struct state *sp;
	register struct client *cl, *nextcl;
	fd_set readfds, writefds, treadfds, twritefds;
	socklen_t csalen;
	struct stat sbuf;
	struct sockaddr_in sa, csa;
	static int on = 1;

	if (argv[0] == NULL)
		prog = "acld";
	else if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	openlog(prog, 0, LOG_DAEMON);

	opterr = 0;
	while ((op = getopt(argc, argv, "c:dfP:v")) != EOF)
		switch (op) {

		case 'c':
			configfn = optarg;
			break;

		case 'd':
			++debug;
			break;

		case 'f':
			++foreground;
			break;

		case 'P':
			pidfile = optarg;
			break;

		case 'v':
			++verbose;
			break;

		default:
			usage();
			/* NOTREACHED */
		}

	if (argc != optind)
		usage();

	if (!foreground && daemon(0, 1) < 0) {
		log(LOG_ERR, "daemon: %s", strerror(errno));
		exit(EX_OSERR);
	}

	/* Update pidfile */
	fp = fopen(pidfile, "w");
	if (fp == NULL)
		log(LOG_ERR, "daemon: %s: %s", pidfile, strerror(errno));
	else {
		(void)fprintf(fp, "%d\n", (int)getpid());
		(void)fclose(fp);
	}

	log(LOG_INFO, "version %s starting", version);

	(void)setsignal(SIGTERM, sigterm);
	(void)setsignal(SIGINT, sigterm);
	(void)setsignal(SIGHUP, sighup);
	(void)setsignal(SIGCHLD, sigchild);
	(void)setsignal(SIGPIPE, SIG_IGN);

	sp = &state;
	sp->state = ASTATE_NOTCONNECTED;
	sp->rfd = -1;
	sp->wfd = -1;
	sp->s = -1;
	reload = 1;
	sp->nfds = 0;
	memset(&sa, 0, sizeof(sa));
#ifdef HAVE_SOCKADDR_SA_LEN
	sa.sin_len = sizeof(sa);
#endif
	sa.sin_family = AF_INET;
	for (;;) {
		if (sawterm) {
			/* Terminate the child */
			log(LOG_NOTICE, "exiting");
			/* simulate /bin/sh exit status */
			exit(128 + SIGTERM);
		}
		if (reload) {
			/* Terminate the child */
			if (sp->wfd >= 0 && sp->state == ASTATE_LOGGEDIN)
				childkill(sp);

			/* Reload config file if it's changed */
			if (sp->cf != NULL) {
				/* Get the config file mod time */
				if (stat(configfn, &sbuf) < 0)
					log(LOG_ERR, "stat %s: %s",
					    configfn, strerror(errno));
				else {
					t = (sbuf.st_mtime > sbuf.st_ctime) ?
					    sbuf.st_mtime : sbuf.st_ctime;
					if (t > sp->cf->c_time) {
						log(LOG_INFO, "reloading %s",
						    configfn);
						freecf(sp->cf);
						sp->cf = NULL;
					}
				}
			}
			reload = 0;
			sp->nfds = 0;
			timerset(&sp->t_login, 1);
		}

		/* Read the config file */
		if (sp->cf == NULL) {
			if ((sp->cf = parsecf(configfn)) == NULL)
				exit(EX_CONFIG);
			sa.sin_port = htons(sp->cf->c_port);
			sa.sin_addr.s_addr = sp->cf->c_bindaddr;
		}

		/* Create new expect process */
		if (sp->wfd < 0) {
			if (expect(sp) < 0) {
				log(LOG_ERR, "expect child: %s",
				    strerror(errno));
				exit(EX_OSERR);
			}
			sp->nfds = 0;
		}

		/* Internal processing */
		if (sp->wfd >= 0) {
			switch (sp->state) {

			case ASTATE_CONNECTED:
				if (timerdue(&sp->t_login))
					childlogin(sp);
				break;

			case ASTATE_LOGGEDIN:
				if (!sp->listedallacls)
					childlistacl(sp);
				break;

			default:
				break;
			}
		}

		/* Open the listen socket */
		if (sp->s < 0) {
			sp->s = socket(PF_INET, SOCK_STREAM, 0);
			if (sp->s < 0) {
				log(LOG_ERR, "socket: %s", strerror(errno));
				exit(EX_OSERR);
			}
			if (setsockopt(sp->s, SOL_SOCKET, SO_REUSEADDR,
			    (char *)&on, sizeof (on)) < 0) {
				log(LOG_ERR, "SO_REUSEADDR: %s",
				    strerror(errno));
				exit(EX_OSERR);
			}
			if (bind(sp->s,
			    (struct sockaddr *)&sa, sizeof(sa)) < 0) {
				log(LOG_ERR, "bind: %s", strerror(errno));
				exit(EX_OSERR);
			}
			if (listen(sp->s, backlog) < 0) {
				log(LOG_ERR, "listen: %s", strerror(errno));
				exit(EX_OSERR);
			}
		}

		/* Update prototype fd_sets */
		if (sp->nfds <= 0) {
			FD_ZERO(&treadfds);
			FD_ZERO(&twritefds);
			sp->nfds = sp->s;
			FD_SET(sp->s, &treadfds);
			if (sp->rfd >= 0) {
				FD_SET(sp->rfd, &treadfds);
				sp->nfds = MAX(sp->nfds, sp->rfd);
			}
			for (cl = sp->clients; cl != NULL; cl = cl->next) {
				FD_SET(cl->c, &treadfds);
				sp->nfds = MAX(sp->nfds, cl->c);

				if (cl->wbuf.len > 0)
					FD_SET(cl->c, &twritefds);
			}
			++sp->nfds;
		}

		readfds = treadfds;
		writefds = twritefds;

		/* Wait for stuff to do (or timeout) */
		settimeout(sp);
		n = select(sp->nfds, &readfds, &writefds, NULL, &sp->timeout);
		if (n < 0) {
			/* Don't choke if we get ptraced */
			if (errno == EINTR)
				continue;
			log(LOG_ERR, "select: %s", strerror(errno));
			exit(EX_OSERR);
		}

		/* Time for maintenance */
		if (n == 0) {
			/* Did the child fall asleep? */
			if (sp->wfd < 0 || sp->state != ASTATE_LOGGEDIN) {
				if (timerdue(&sp->t_ayt))
					childkill(sp);
				continue;
			}

			/* Only do maintenance if there's nothing else to do */
			if (sp->req != NULL || sp->reqclient != NULL)
				continue;

			if (timerdue(&sp->t_compact))
				servercompact(sp);
			else if (timerdue(&sp->t_sync))
				serversync(sp);
			else if (timerdue(&sp->t_ayt))
				serverayt(sp);
		}

		/* Did any clients say something? */
		for (cl = sp->clients; cl != NULL; cl = nextcl) {
			/* Ok to free client inside this loop */
			nextcl = cl->next;
			if (cl->c >= 0 && FD_ISSET(cl->c, &readfds)) {
				cc = ioread(cl->c, &cl->rbuf);
				if (cc < 0) {
					log(LOG_ERR, "client #%d read: %s",
					    cl->n, strerror(errno));
					freeclient(sp, cl);
					continue;
				}
				if (cc == 0) {
					log(LOG_INFO,
					    "client #%d disconnected", cl->n);
					freeclient(sp, cl);
					continue;
				}
			}
		}

		/* Check for a new client */
		if (sp->s >= 0 && FD_ISSET(sp->s, &readfds)) {
			memset(&csa, 0, sizeof(csa));
			csalen = sizeof(csa);
			c = accept(sp->s, (struct sockaddr *)&csa, &csalen);
			if (c < 0) {
				log(LOG_ERR, "accept: %s", strerror(errno));
				if (errno == EBADF)
					exit(EX_OSERR);
				continue;
			}
			nbio(c);
			cl = newclient(sp, c);
			log(LOG_DEBUG, "client #%d connect from %s.%hu",
			    cl->n, inet_ntoa(csa.sin_addr), csa.sin_port);

			clientsend(sp, cl, NULL, CS_CONTINUE);
			clientwrite(sp, cl, "ready for action");
			clientsendeom(sp, cl);
		}

		/* Did the child say something? */
		if (sp->rfd >= 0 && FD_ISSET(sp->rfd, &readfds)) {
			cc = ioread(sp->rfd, &sp->rbuf);
			if (cc < 0) {
				log(LOG_ERR, "child read: %s", strerror(errno));
				continue;
			}
			if (cc == 0) {
				log(LOG_DEBUG, "child disconnect");
				sp->state = ASTATE_NOTCONNECTED;
				sp->cmd = NULL;
				childkill(sp);
				continue;
			}
		}

		/* Process child input */
		while (sp->rbuf.len > 0) {
			cl = sp->reqclient;
			childinput(sp, cl);

			/* Remove server request from list if completed */
			if (sp->req != NULL && sp->req->state == RSTATE_DONE) {
				/* Remove from request list */
				rp = sp->req;
				sp->req = rp->next;
				rp->next = NULL;
				freereq(rp);
				continue;
			}

			/* Remove client request from list if completed */
			if (cl != NULL && cl->req != NULL &&
			    cl->req->state == RSTATE_DONE) {
				/* Remove from request list */
				rp = cl->req;
				cl->req = rp->next;
				rp->next = NULL;
				freereq(rp);
				sp->reqclient = NULL;
				continue;
			}
		}

		/*
		 * Process input from each client input, possibly
		 * appending a request to the end of the client's
		 * queue.
		 */
		for (cl = sp->clients; cl != NULL; cl = cl->next)
			while (iohaveline(&cl->rbuf))
				clientinput(sp, cl);

		/* Can't process client requests until we get logged in */
		if (sp->state != ASTATE_LOGGEDIN)
			continue;

		/* Reset keepalive if necessary */
		if (sp->cf->c_ayt_secs > 0 && timercheck(&sp->t_ayt) < 0)
			timerset(&sp->t_ayt, sp->cf->c_ayt_secs);

		/* Handle server requests */
		while ((rp = sp->req) != NULL && rp->state == RSTATE_PENDING) {
			serverprocess(sp, rp);
			if (rp->state == RSTATE_DONE) {
				/* Remove from request list */
				rp = sp->req;
				sp->req = rp->next;
				rp->next = NULL;
				freereq(rp);
			}
		}

		/* Handle client requests */
		/* XXX try to give clients a fair chance? */
		if (sp->req == NULL)
		for (cl = sp->clients; cl != NULL; cl = cl->next)
			while (sp->reqclient == NULL &&
			    cl->req != NULL &&
			    cl->req->state == RSTATE_PENDING) {
				sp->reqclient = cl;
				clientprocess(sp, cl, cl->req);
				if (cl->req->state == RSTATE_DONE) {
					/* Remove from request list */
					rp = cl->req;
					cl->req = rp->next;
					rp->next = NULL;
					freereq(rp);
					sp->reqclient = NULL;
				}
			}

		/* Drain client output */
		for (cl = sp->clients; cl != NULL; cl = nextcl) {
			/* Ok to free client inside this loop */
			nextcl = cl->next;
			if (FD_ISSET(cl->c, &writefds) && cl->wbuf.len > 0) {
				if (iowrite(cl->c, &cl->wbuf) < 0) {
					log(LOG_ERR, "client #%d write: %s",
					    cl->n, strerror(errno));
					freeclient(sp, cl);
					continue;
				}
				/* If this client is done free him */
				if (cl->close && cl->wbuf.len <= 0) {
					freeclient(sp, cl);
					continue;
				}
				/* When finished writing redo fd_sets */
				if (cl->wbuf.len <= 0)
					sp->nfds = 0;
			}
		}
	}
}

void
intlistadd(register struct cf *cf, register const char *str)
{
	register struct intlist *ip;
	register int an, i;
	char **av;
	char buf[1024];
	register char *cp;
	size_t size, len;

	an = makeargv(str, &av);
	if (an < 2) {
		log(LOG_ERR, "intadd wrong number of args: \"%s\"", str);
		freeargv(av);
		return;
	}

	dynarray((void **)&cf->c_intlist, &cf->c_intlistlen,
	    &cf->c_intlistsize, sizeof(*cf->c_intlist),
	    1, 1, "cf intlist");
	ip = cf->c_intlist + cf->c_intlistlen;
	++cf->c_intlistlen;

	ip->i_acl = strsave(av[an - 1]);

	size = sizeof(buf);
	cp = buf;
	for (i = 0; i < an - 1; ++i) {
		(void)snprintf(cp, size, " %s", av[i]);
		len = strlen(cp);
		cp += len;
		size -= len;
	}
	cp = buf + 1;
	ip->i_name = strsave(cp);

	freeargv(av);
}

void
intlistfree(register struct intlist *ip)
{

	if (ip->i_acl != NULL) {
		free(ip->i_acl);
		ip->i_acl = NULL;
	}
	if (ip->i_name != NULL) {
		free(ip->i_name);
		ip->i_name = NULL;
	}
}

/* Create the expect process */
int
expect(register struct state *sp)
{
	register pid_t pid;
	register int xerrno;
	int wfds[2], rfds[2];
	char *argv[3], *envp[1];

	if (access(sp->cf->c_script, R_OK) < 0) {
		log(LOG_ERR, "expect: %s: %s",
		    sp->cf->c_script, strerror(errno));
		exit(EX_OSFILE);
	}

	if (access(sp->cf->c_expect, X_OK) < 0) {
		log(LOG_ERR, "expect: %s: %s",
		    sp->cf->c_expect, strerror(errno));
		exit(EX_OSFILE);
	}

	if (pipe(rfds) < 0)
		return (-1);
	if (pipe(wfds) < 0) {
		xerrno = errno;
		(void)close(rfds[0]);
		(void)close(rfds[1]);
		errno = xerrno;
		return (-1);
	}

	pid = fork();
	if (pid < 0) {
		xerrno = errno;
		(void)close(rfds[0]);
		(void)close(rfds[1]);
		(void)close(wfds[0]);
		(void)close(wfds[1]);
		errno = xerrno;
		return (-1);
	}
	if (pid == 0) {
		/* Child */
		/* XXX close all descriptors?? */
		/* Setup stdin, stdout and stderr */
		if (rfds[0] != STDIN_FILENO) {
			(void)dup2(rfds[0], STDIN_FILENO);
			(void)close(rfds[0]);
		}
		(void)close(rfds[1]);

		if (wfds[1] != STDOUT_FILENO) {
			(void)dup2(wfds[1], STDOUT_FILENO);
			(void)close(wfds[1]);
		}
		(void)dup2(STDOUT_FILENO, STDERR_FILENO);
		(void)close(wfds[0]);

		argv[0] = sp->cf->c_expect;
		argv[1] = sp->cf->c_script;
		argv[2] = NULL;
		envp[0] = NULL;
		(void)execve(sp->cf->c_expect, argv, envp);
		log(LOG_ERR, "execve: %s", strerror(errno));
		exit(EX_OSERR);
	}


	/* Parent */
	(void)close(rfds[0]);
	(void)close(wfds[1]);

	/* The child's write is our read, etc. */
	sp->rfd = wfds[0];
	sp->wfd = rfds[1];
	sp->pid = pid;
	log(LOG_DEBUG, "expect: spawed child %d", (int)pid);
	return (0);
}

/* Append the new request to end of list */
void
appendreq(register struct req **reqp, register struct req *rp)
{
	register struct req *rp2;

	if (*reqp == NULL)
		*reqp = rp;
	else {
		rp2 = *reqp;
		while (rp2->next != NULL)
			rp2 = rp2->next;
		rp2->next = rp;
	}
}

void
freereq(struct req *rp)
{
	register int i;
	register struct state *sp;
	register struct acllist *ap;

	if (rp == NULL)
		return;

	if (rp->next != NULL) {
		log(LOG_ERR, "freereq: rp->next not NULL");
		exit(EX_SOFTWARE);
	}
	if (rp->aclname != NULL)
		free(rp->aclname);
	if (rp->comment != NULL)
		free(rp->comment);
	if (rp->av != NULL)
		freeargv(rp->av);
	free(rp);

	/* Clean up compaction state */
	sp = &state;		/* XXX */
	for (i = 0, ap = sp->cf->c_acllist; i < sp->cf->c_acllistlen; ++i, ++ap)
		if (ap->compactreq == rp) {
			ap->compactclient = NULL;
			ap->compactreq = NULL;
			break;
		}
}


void
nbio(register int s)
{
	register int flags;

	if ((flags = fcntl(s, F_GETFL, 0)) < 0) {
		log(LOG_ERR, "nbio: F_GETFL: %s", strerror(errno));
		return;
	}
	flags |= O_NONBLOCK;
	if ((flags = fcntl(s, F_SETFL, flags)) < 0) {
		log(LOG_ERR, "nbio: F_SETFL: %s", strerror(errno));
		return;
	}
}

/* Set our select timeout */
void
settimeout(register struct state *sp)
{
	register int t, secs;

	secs = sp->cf->c_select_secs;
	t = timercheck(&sp->t_login);
	if (t >= 0 && secs > t)
		secs = t;
	t = timercheck(&sp->t_sync);
	if (t >= 0 && secs > t)
		secs = t;
	t = timercheck(&sp->t_compact);
	if (t >= 0 && secs > t)
		secs = t;
	t = timercheck(&sp->t_ayt);
	if (t >= 0 && secs > t)
		secs = t;

	/* Always sleep for at least 10ms */
	if (secs > 0) {
		sp->timeout.tv_sec = secs;
		sp->timeout.tv_usec = 0;
	} else {
		sp->timeout.tv_sec = 0;
		sp->timeout.tv_usec = 10 * 1000;
	}

}

/* Reap exiting child */
RETSIGTYPE
sigchild(int signo)
{
	register pid_t pid;
	DECLWAITSTATUS status;

	pid = waitpid(0, &status, WNOHANG);
	return RETSIGVAL;
}

RETSIGTYPE
sighup(int signo)
{

	++reload;
	return RETSIGVAL;
}

#ifdef notdef
RETSIGTYPE
sigpipe(int signo)
{
	register struct state *sp;

	sp = &state;
	if (sp->c >= 0) {
		close(sp->c);
		sp->c = -1;
		sp->nfds = 0;
	}

	return RETSIGVAL;
}
#endif

RETSIGTYPE
sigterm(int signo)
{
	register struct state *sp;
	register struct client *cl;

	sp = &state;
	for (cl = sp->clients; cl != NULL; cl = cl->next) {
		close(cl->c);
		cl->c = -1;
	}
	sp->nfds = 0;

	++sawterm;
	return RETSIGVAL;
}

__dead void
usage(void)
{

	(void)fprintf(stderr, "%s version %s\n", prog, version);
	(void)fprintf(stderr, "Usage: %s [-dfv] [-c config] [-P pidfile]\n",
	    prog);
	exit(EX_USAGE);
}
