/*
 * Copyright (c) 2002, 2003, 2004
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: child.c,v 1.35 2004/09/13 17:51:55 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "cf.h"
#include "child.h"
#include "util.h"

/* Forwards */
int childresponse(char **, struct state *);
int failed(const char *, const char *);
static void nets_log(struct state *, const char *, struct req *);

static void
nets_log(register struct state *sp, register const char *what,
    register struct req *rp)
{
	const char *p, *w;

	if (rp->comment == NULL)
		p = "";
	else
		p = pretty(rp->comment, strlen(rp->comment));

	switch (rp->acl.type) {

	case ATYPE_PERMITHOST:
	case ATYPE_BLOCKHOST:
		w = addr2str(rp->acl.addr1.addr);
		break;

	default:
		w = "?";
		break;
	}
	syslog(sp->cf->c_netsfac | LOG_DEBUG,
	    "NETS status=%s ats=%ld.%06ld cts=%ld.%06ld cmd=%s ip=%s cmt=%s",
	    what,
	    rp->ats.tv_sec, rp->ats.tv_usec,
	    rp->cts.tv_sec, rp->cts.tv_usec,
	    rp->cmd,
	    w,
	    p);
}

/* Process input from the client */
void
childinput(register struct state *sp, register struct client *cl)
{
	register int an, flags, i;
	register char *cp;
	register struct acllist *ap, *ap2;
	register struct req *rp;
	char **av;
	static char *p1, *p2;
	register struct iobuf *ip;

	if (cl != NULL)
		rp = cl->req;
	else
		rp = sp->req;

	ip = &sp->rbuf;
	switch (sp->state) {

	case ASTATE_READRESPONSE:
		if (childresponse(&p1, sp)) {
			if (sp->state == ASTATE_READRESPONSE) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
			} else {
				sp->state = ASTATE_CONNECTED;
				sp->cmd = NULL;
			}
			if (p1 != NULL) {
				trimws(p1);
				log(LOG_DEBUG, "childinput response: \"%s\"",
				    pretty(p1, strlen(p1)));
				free(p1);
				p1 = NULL;
			}
		}
		return;

	case ASTATE_READERROR:
		if (childresponse(&p2, sp)) {
			if (sp->state == ASTATE_READRESPONSE) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
			} else {
				sp->state = ASTATE_CONNECTED;
				sp->cmd = NULL;
				/* Wait a bit before trying again */
				timerset(&sp->t_login, sp->cf->c_login_secs);
			}
			if (p2 != NULL) {
				trimws(p2);
				log(LOG_DEBUG, "childinput error: \"%s\"",
				    pretty(p2, strlen(p2)));
				free(p2);
				p2 = NULL;
			}
		}
		return;

	case ASTATE_READACL:
		while ((cp = iogetstr(ip)) != NULL) {
			/* Find the last acl list that hasn't been updated */
			ap2 = sp->cf->c_acllist;
			ap = NULL;
			for (i = 0; i < sp->cf->c_acllistlen; ++i) {
				if (!ap2->sentlistacl)
					break;
				ap = ap2++;
			}
			if (ap == NULL) {
				log(LOG_ERR, "childinput: can't happen");
				exit(EX_SOFTWARE);
			}

			/* Remove trailing newline for possible error message */
			trimws(cp);
			if (aclstraddacl(sp->cf, ap, cp)) {
				sp->state = ASTATE_LOGGEDIN;
				sp->cmd = NULL;
				break;
			}
		}
		return;

	case ASTATE_NOTCONNECTED:
	case ASTATE_SENTLOGIN:
	case ASTATE_LOGGEDIN:
	case ASTATE_SENTLISTACL:
		break;

	default:
		if (rp != NULL && rp->state == RSTATE_READRESPONSE) {
			while ((cp = iogetstr(ip)) != NULL) {
				/* Insure proper trailing "\r\n" */
				trimws(cp);
				if (strcmp(cp, ".") == 0) {
					/* XXX? */
					if (cl != NULL)
						clientsendeom(sp, cl);
					if (rp != NULL) {
						rp->state = RSTATE_DONE;
						getts(&rp->cts);
					}
					return;
				}
				log(LOG_DEBUG, "childinput response: \"%s\"",
				    pretty(cp, strlen(cp)));
				/* XXX debug log if it's for the server? */
				if (cl != NULL)
					clientwrite(sp, cl, "%s", cp);
			}
			return;
		}
		break;
	}

	/* Break into arguments */
	cp = iogetstr(ip);
	av = NULL;
	an = makeargv(cp, &av);

	if (an <= 0) {
		freeargv(av);
		return;
	}

	if (an == 1 && strcmp(av[0], "expect>") == 0) {
		if (sp->state == ASTATE_NOTCONNECTED) {
			sp->state = ASTATE_CONNECTED;
			sp->cmd = NULL;
		}
		if (debug)
			fprintf(stderr, "%s childinput: (ready)\n", tsstr());
		freeargv(av);
		return;
	}

	if (debug) {
		fprintf(stderr, "%s childinput:", tsstr());
		for (i = 0; i < an; ++i)
			fprintf(stderr, " %s", av[i]);
		putc('\n', stderr);
	}

	switch (sp->state) {

	case ASTATE_SENTLISTACL:
	case ASTATE_SENTLOGIN:
		flags = 0;
		if (failed(sp->cmd, av[0]))
			flags |= CS_FAILED;
		else if (strcmp(sp->cmd, av[0]) != 0) {
			log(LOG_ERR,
			    "childinput: unexpected response %s waiting for %s",
			    av[0], sp->cmd);
			freeargv(av);
			return;
		}
		if (an > 1 && strcmp(av[1], "-") == 0)
			flags |= CS_CONTINUE;

		switch (sp->state) {

		case ASTATE_SENTLOGIN:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READRESPONSE;
			} else {
				if ((flags & CS_CONTINUE) == 0) {
					sp->state = ASTATE_CONNECTED;
					/* Wait a bit before trying again */
					timerset(&sp->t_login,
					    sp->cf->c_login_secs);
				} else
					sp->state = ASTATE_READERROR;
			}
			break;

		case ASTATE_SENTLISTACL:
			if ((flags & CS_FAILED) == 0) {
				if ((flags & CS_CONTINUE) == 0)
					sp->state = ASTATE_LOGGEDIN;
				else
					sp->state = ASTATE_READACL;
			} else {
				if ((flags & CS_CONTINUE) == 0) {
					sp->state = ASTATE_LOGGEDIN;
					childkill(sp);
				} else
					sp->state = ASTATE_READERROR;
			}
			break;

		default:
			break;
		}

		freeargv(av);
		return;

	default:
		break;
	}

	if (rp == NULL) {
		log(LOG_ERR,
		    "childinput: no server/client request to process");
		freeargv(av);
		return;
	}

	if (rp->state != RSTATE_CHILD) {
		log(LOG_ERR, "childinput: not waiting for child to talk!");
		freeargv(av);
		return;
	}

	flags = 0;
	if (failed(rp->cmd, av[0]))
		flags = CS_FAILED;
	else if (strcmp(rp->cmd, av[0]) != 0) {
		log(LOG_ERR,
		    "childinput: unexpected response %s waiting for %s",
		    av[0], rp->cmd);
		freeargv(av);
		return;
	}

	if (an > 1 && strcmp(av[1], "-") == 0) {
		flags |= CS_CONTINUE;
		rp->state = RSTATE_READRESPONSE;
	} else {
		rp->state = RSTATE_DONE;
		getts(&rp->cts);

		/* log for NETS */
		if (rp->type == REQ_DROP || rp->type == REQ_RESTORE)
			nets_log(sp, "success", rp);
	}

	if (cl != NULL)
		clientsend(sp, cl, rp, flags);

	/* Fix up things if updating the router failed */
	if ((flags & CS_FAILED) != 0) {
		switch (rp->type) {

		case REQ_DROP:
		case REQ_RESTORE:
			/* Need to delete/restore acl entry since we failed */
			ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
			if (ap == NULL) {
				log(LOG_ERR, "childinput: NULL %s", rp->cmd);
				exit(EX_SOFTWARE);
			}
			if (rp->type == REQ_DROP) {
				if (!acldeleteacl(ap, &rp->acl))
					log(LOG_ERR,
					    "childinput: acldeleteacl failed");
			} else
				acladdacl(ap, &rp->acl);
			break;

		case REQ_DROPUDPPORT:
		case REQ_DROPTCPPORT:
		case REQ_RESTOREUDPPORT:
		case REQ_RESTORETCPPORT:
			/* Need to delete/restore acl entry since we failed */
			ap = aclfindlistbyname(sp->cf, rp->aclname);
			if (ap == NULL) {
				log(LOG_ERR, "childinput: NULL %s", rp->cmd);
				exit(EX_SOFTWARE);
			}
			if (rp->type == REQ_DROPTCPPORT) {
				if (!acldeleteacl(ap, &rp->acl))
					log(LOG_ERR,
					    "childinput: acldeleteacl failed");
			} else
				acladdacl(ap, &rp->acl);
			break;

		default:
			break;
		}
	}

	freeargv(av);
}

/* XXX unfortunately, this duplicates a lot of serverdroprestore() */
void
childdroprestore(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register struct acllist *ap;
	register struct acl *al;
	char buf[132];

	/* Find acl list */
	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
		ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
		if (ap == NULL) {
			clientsenderr(sp, cl, rp,
			    "can't find acllist for %s"
			    " (is the default acl missing?)",
			    addr2str(rp->acl.addr1.addr));
			rp->state = RSTATE_DONE;
			return;
		}
		break;

	case ATYPE_BLOCKUDPPORT:
	case ATYPE_BLOCKTCPPORT:
		ap = aclfindlistbyname(sp->cf, rp->aclname);
		if (ap == NULL) {
			clientsenderr(sp, cl, rp,
			    "can't find acllist for %s"
			    " (is the default acl missing?)",
			    rp->acl.port1);
			rp->state = RSTATE_DONE;
			return;
		}
		break;

	default:
		clientsenderr(sp, cl, rp,
		    "Acl type %d not implemented", rp->acl.type);
		rp->state = RSTATE_DONE;
		return;
	}

	/* Check if acl should or should not already exist */
	al = aclfindacl(ap, &rp->acl);
	switch (rp->type) {

	case REQ_DROP:
		if (al != NULL) {
			clientsend(sp, cl, rp, CS_CONTINUE);
			clientwrite(sp, cl,
			    "Note: %s is already blocked in acl %s",
			    addr2str(rp->acl.addr1.addr), ap->name);
			clientsendeom(sp, cl);
			rp->state = RSTATE_DONE;

			/* log for NETS */
			nets_log(sp, "ignore", rp);
			return;
		}
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
		if (al != NULL) {
			clientsend(sp, cl, rp, CS_CONTINUE);
			clientwrite(sp, cl,
			    "Note: %s is already blocked in acl %s",
			    rp->acl.port1, ap->name);
			clientsendeom(sp, cl);
			rp->state = RSTATE_DONE;
			return;
		}
		break;

	case REQ_RESTORE:
		if (al == NULL) {
			clientsenderr(sp, cl, rp,
			    "%s not found in acl %s",
			    addr2str(rp->acl.addr1.addr), ap->name);
			rp->state = RSTATE_DONE;

			/* log for NETS */
			nets_log(sp, "ignore", rp);
			return;
		}
		break;


	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
		if (al == NULL) {
			clientsenderr(sp, cl, rp,
			    "%s not found in acl %s", rp->acl.port1, ap->name);
			rp->state = RSTATE_DONE;
			return;
		}
		break;

	default:
		log(LOG_ERR, "childdroprestore: bad type %d (1)", rp->type);
		exit(EX_SOFTWARE);
	}

	/* Add or remove the acl */
	switch (rp->type) {

	case REQ_DROP:
		/* Are we moving up or down? */
		if (sp->cf->c_incrseq) {
			if (sp->cf->c_highseq > 0 &&
			    ap->lastseq >= sp->cf->c_highseq) {
				snprintf(buf, sizeof(buf),
				    "Last sequence number in ACL %s"
				    " already in use (%d)",
				    ap->name, sp->cf->c_highseq);
				clientsenderr(sp, cl, rp, buf);
				syslog(LOG_INFO, buf);
				rp->state = RSTATE_DONE;
				return;
			}
		} else {
			if (sp->cf->c_lowseq > 0 &&
			    ap->lastseq <= sp->cf->c_lowseq) {
				snprintf(buf, sizeof(buf),
				    "First sequence number in ACL %s"
				    " already in use (%d)",
				    ap->name, sp->cf->c_lowseq);
				clientsenderr(sp, cl, rp, buf);
				syslog(LOG_INFO, buf);
				rp->state = RSTATE_DONE;
				return;
			}
		}

		/* Check cap on maxium number of acls */
		if (sp->cf->c_maxseq > 0 && ap->acllen >= sp->cf->c_maxseq) {
			snprintf(buf, sizeof(buf),
			    "No free slots in ACL %s (%d in use, %d allowed)",
			    ap->name, ap->acllen, sp->cf->c_maxseq);
			clientsenderr(sp, cl, rp, buf);
			syslog(LOG_INFO, buf);
			rp->state = RSTATE_DONE;
			return;
		}

		if (sp->cf->c_incrseq)
			rp->acl.seq = ++ap->lastseq;
		else
			rp->acl.seq = --ap->lastseq;

		/* Add here to avoid the race (remove later if it fails) */
		acladdacl(ap, &rp->acl);
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
		if (sp->cf->c_highportseq > 0 &&
		    ap->lastportseq >= sp->cf->c_highportseq) {
			clientsenderr(sp, cl, rp,
			    "Last port sequence number in ACL %s"
			    " already in use (%d)",
			    ap->name, sp->cf->c_highportseq);
			rp->state = RSTATE_DONE;
			return;
		}

		rp->acl.seq = ++ap->lastportseq;

		/* Add here to avoid the race (remove later if it fails) */
		acladdacl(ap, &rp->acl);
		break;

	case REQ_RESTORE:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
		rp->acl.seq = al->seq;

		/* Remove here to avoid the race (restore later if it fails) */
		if (!acldeleteacl(ap, &rp->acl)) {
			log(LOG_ERR, "childdroprestore: acldeleteacl failed");
			exit(EX_SOFTWARE);
		}
		break;

	default:
		log(LOG_ERR, "childdroprestore: bad type %d (2)", rp->type);
		exit(EX_SOFTWARE);
	}

	/* Send the request to the child */
	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
		childsend(sp, "%s %s %s %d",
		    rp->cmd, addr2str(rp->acl.addr1.addr),
		    ap->name, rp->acl.seq);
		break;

	case ATYPE_BLOCKUDPPORT:
	case ATYPE_BLOCKTCPPORT:
		childsend(sp, "%s %s %s %d",
		    rp->cmd, rp->acl.port1, ap->name, rp->acl.seq);
		break;

	default:
		log(LOG_ERR, "childdroprestore: bad atype %d", rp->acl.type);
		exit(EX_SOFTWARE);
	}

	/* Update request state */
	rp->state = RSTATE_CHILD;

	/* Schedule a sync */
	if (sp->cf->c_sync_secs > 0 && timercheck(&sp->t_sync) < 0)
		timerset(&sp->t_sync, sp->cf->c_sync_secs);
}

void
childkill(register struct state *sp)
{

	if (sp->wfd >= 0 && sp->state == ASTATE_LOGGEDIN)
		childsend(sp, "logout");
	sp->state = ASTATE_NOTCONNECTED;
	sp->cmd = NULL;

	errno = 0;
	if (sp->pid > 0 && kill(sp->pid, SIGTERM) < 0 && errno != ESRCH)
		log(LOG_ERR, "childkill: kill %d: %s",
		    (int)sp->pid, strerror(errno));

	sp->pid = -1;
	if (sp->rfd >= 0)
		close(sp->rfd);
	sp->rfd = -1;

	if (sp->wfd >= 0)
		close(sp->wfd);
	sp->wfd = -1;

	sp->nfds = 0;

	iofree(&sp->rbuf);

	timerset(&sp->t_ayt, 0);
	timerset(&sp->t_login, 0);
	timerset(&sp->t_sync, 0);
}

void
childlistacl(register struct state *sp)
{
	register struct acllist *ap;
	register struct cf *cf;
	register int i;

	cf = sp->cf;
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		if (!ap->sentlistacl) {
			sp->state = ASTATE_SENTLISTACL;
			sp->cmd = "listacl";
			childsend(sp, "%s %s {%s}",
			    sp->cmd, ap->name, ap->intlist->i_name);
			++ap->sentlistacl;
			return;
		}

	/* If we got here, we've listed all the ACLs; sort 'em */
	++sp->listedallacls;
	aclsortacls(cf);
}

void
childlogin(register struct state *sp)
{
	register struct cf *cf;
	int odebug;

	/* Free acl lists so we'll re-aquire them from the router */
	cf = sp->cf;
	acllistsfree(sp);

	sp->state = ASTATE_SENTLOGIN;
	sp->cmd = "login";
	odebug = debug;
	if (odebug) {
		/* Avoid exposing router passwords in debugging output */
		debug = 0;
		fprintf(stderr, "%s childsend: \"%s %s %s %s %s %s %s %s\"\n",
		    tsstr(),
		    sp->cmd,
		    cf->c_router,
		    cf->c_cuser,
		    "?",
		    "?",
		    cf->c_euser,
		    "?",
		    "?");
	}
	childsend(sp, "%s %s %s %s %s %s %s %s",
	    sp->cmd,
	    cf->c_router,
	    cf->c_cuser,
	    cf->c_cpass1,
	    cf->c_cpass2,
	    cf->c_euser,
	    cf->c_epass1,
	    cf->c_epass2);
	debug = odebug;
	timerreset(&sp->t_ayt);
	timerreset(&sp->t_sync);
}

int
childresponse(register char **pp, register struct state *sp)
{
	register char *cp;
	register struct iobuf *ip;

	/* Look for end of message (and strip leading dots) */
	ip = &sp->rbuf;
	while ((cp = iogetstr(ip)) != NULL) {
		if (*cp == '.') {
			++cp;
			if (*cp == '\0' || *cp == '\n')
				return (1);
		}
		strappend(pp, cp);
	}
	return(0);
}

void
childsend(register struct state *sp, register const char *fmt, ...)
{
	register char *cp;
	register size_t n;
	register ssize_t cc;
	char buf[1024];
	va_list ap;

	va_start(ap, fmt);
	(void)vsnprintf(buf, sizeof(buf) - 1, fmt, ap);
	va_end(ap);
	n = strlen(buf);
	cp = buf + n - 1;

	if (debug)
		fprintf(stderr, "%s childsend: \"%s\"\n", tsstr(), buf);

	if (cp >= buf) {
		*++cp = '\n';
		*++cp = '\0';
		++n;
	}
	cc = write(sp->wfd, buf, n);
	if (cc < 0) {
		/* XXX close off child here? */
		log(LOG_ERR, "childsend: write: %s", strerror(errno));
		return;
	}
	if (cc != n) {
		log(LOG_ERR, "childsend: short write (%d != %d)", cc, n);
		return;
	}
}

int
failed(register const char *what, register const char *str)
{
	register size_t len;

	len = strlen(what);
	if (strncmp(what, str, len) == 0 && strcmp(str + len, "-failed") == 0)
		return (1);
	return (0);
}
