/*
 * Copyright (c) 2002, 2003, 2004
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: client.c,v 1.34 2004/09/13 17:51:47 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "cf.h"
#include "child.h"
#include "util.h"

/* Globals */
struct s2v cmd2req[] = {
	{ "reload",	REQ_RELOAD },
	{ "drop",	REQ_DROP },
	{ "restore",	REQ_RESTORE },
	{ "dropudpport", REQ_DROPUDPPORT },
	{ "restoreudpport", REQ_RESTOREUDPPORT },
	{ "droptcpport", REQ_DROPTCPPORT },
	{ "restoretcpport", REQ_RESTORETCPPORT },
	{ "state",	REQ_STATE },
	{ "listacl",	REQ_LISTACL },
	{ "query",	REQ_QUERY },
	{ "compact",	REQ_COMPACT },
	{ "exit",	REQ_EXIT },
	{ "test",	REQ_TEST },		/* undocumented */
	{ "help",	REQ_HELP },
	{ "unknown",	REQ_UNKNOWN },
	{ NULL,		REQ_UNKNOWN }
};

/* Forwards */
int clientcomment(struct req *, struct iobuf *);
void clientcompact(struct state *, struct client *, struct req *);
void clienthelp(struct state *, struct client *, struct req *);
void clientlistacl(struct state *, struct client *, struct req *);
void clientquery(struct state *, struct client *, struct req *);
void clientstate(struct state *, struct client *, struct req *);
void clientwritevar(struct state *, struct client *, const char *, va_list);

/* Read a command from the client */
void
clientinput(register struct state *sp, register struct client *cl)
{
	register char *cp;
	register const char *p;
	register int i;
	char *ep;
	struct req *rp;

	rp = cl->treq;
	if (rp == NULL) {
		/* New request */
		rp = new(1, sizeof(*rp), "clientinput: req");
		cl->treq = rp;
		rp->state = RSTATE_PENDING;

		/* Arrival timestamp */
		getts(&rp->ats);

		cp = iogetstr(&cl->rbuf);
		rp->an = makeargv(cp, &rp->av);

		/* Do we have a comment? */
		if (rp->an > 2 && strcmp(rp->av[rp->an - 1], "-") == 0) {
			--rp->an;
			rp->state = RSTATE_READCOMMENT;

			/* That's all we can do until we collect the comment */
			return;
		}
		/* Fall through and do syntax checks */
	} else {
		/* Request in progress */
		if (rp->state != RSTATE_READCOMMENT) {
			log(LOG_ERR, "clientinput: unexpected input");
			exit(EX_SOFTWARE);
		}

		/* Collect lines until he's done */
		if (!clientcomment(rp, &cl->rbuf))
			return;

		/* Done reading the comment */
		if (rp->comment != NULL) {
			trimws(rp->comment);
			log(LOG_DEBUG, "client #%d comment: \"%s\"",
			    cl->n, pretty(rp->comment, strlen(rp->comment)));
		}
		rp->state = RSTATE_PENDING;
	}

	/* Partial request is now complete */
	cl->treq = NULL;

	if (debug) {
		fprintf(stderr, "%s clientinput: client #%d: \"%s",
		    tsstr(), cl->n, rp->an > 0 ? rp->av[0] : "");
		for (i = 1; i < rp->an; ++i)
			fprintf(stderr, " %s", rp->av[i]);
		fprintf(stderr, "\"\n");
	}

	/* Only say something back if there was a command */
	if (rp->an < 1) {
		freereq(rp);
		return;
	}

	/* Parse command */
	rp->type = str2val(cmd2req, rp->av[0]);
	rp->cmd = val2str(cmd2req, rp->type);
	if (rp->type == REQ_UNKNOWN) {
		clientsenderr(sp, cl, rp,
		    "unknown command \"%s\"", rp->av[0]);
		freereq(rp);
		return;
	}

	/* We need a minimum of a command and a cookie */
	if (rp->an < 2) {
		clientsenderr(sp, cl, rp, "missing cookie");
		freereq(rp);
		return;
	}

	/* Parse cookie */
	rp->cookie = strtol(rp->av[1], &ep, 10);
	if (*ep != '\0') {
		clientsenderr(sp, cl, rp, "non-numeric cookie");
		freereq(rp);
		return;
	}

	/* Verify parameters and build request */
	switch (rp->type) {

	case REQ_RELOAD:
	case REQ_STATE:
	case REQ_EXIT:
	case REQ_HELP:
	case REQ_TEST:
		if (rp->an != 2) {
			clientsenderr(sp, cl, rp, "wrong number of arguments");
			freereq(rp);
			rp = NULL;
			break;
		}
		break;

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_QUERY:
		if (rp->an < 3 || rp->an > 4) {
			clientsenderr(sp, cl, rp, "wrong number of arguments");
			freereq(rp);
			rp = NULL;
			break;
		}

		/* Extract addr */
		if (rp->an == 4)
			cp = rp->av[3];
		else
			cp = NULL;
		p = extractaddr(rp->av[2], cp, &rp->acl.addr1);
		if (p != NULL) {
			clientsenderr(sp, cl, rp, "extractaddr: %s", p);
			freereq(rp);
			rp = NULL;
			break;
		}

		/* XXX This is all we know how to do right now */
		if (rp->acl.addr1.mask == 0xffffffff)
			rp->acl.type = ATYPE_BLOCKHOST;
		break;

	case REQ_DROPUDPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
		if (rp->an != 4) {
			clientsenderr(sp, cl, rp, "wrong number of arguments");
			freereq(rp);
			rp = NULL;
			break;
		}

		if (sp->cf->c_lowportseq == 0 || sp->cf->c_highportseq == 0) {
			clientsenderr(sp, cl, rp,
			    "lowportseq and highportseq are not configured");
			freereq(rp);
			rp = NULL;
			break;
		}

		rp->acl.port1 = strsave(rp->av[2]);
		rp->aclname = strsave(rp->av[3]);
		if (rp->type == REQ_DROPUDPPORT ||
		    rp->type == REQ_RESTOREUDPPORT)
			rp->acl.type = ATYPE_BLOCKUDPPORT;
		else
			rp->acl.type = ATYPE_BLOCKTCPPORT;
		break;

	case REQ_LISTACL:
	case REQ_COMPACT:
		if (rp->an != 3) {
			clientsenderr(sp, cl, rp, "wrong number of arguments");
			freereq(rp);
			rp = NULL;
			break;
		}
		rp->aclname = strsave(rp->av[2]);
		break;

	default:
		break;
	}

	/* Append the new request to end of list */
	if (rp != NULL)
		appendreq(&cl->req, rp);
}

int
clientcomment(register struct req *rp, register struct iobuf *ip)
{
	register char *cp;

	/* Look for end of message (and strip leading dots) */
	while (iohaveline(ip) && (cp = iogetstr(ip)) != NULL) {
		if (*cp == '.') {
			++cp;
			if (strcmp("\r\n", cp) == 0 ||
			    strcmp("\n", cp) == 0 ||
			    *cp == '\0')
				return (1);
		}
		strappend(&rp->comment, cp);
	}
	return (0);
}

void
clientcompact(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register struct acllist *ap;
	register struct req *rp2;

	ap = aclfindlistbyname(sp->cf, rp->aclname);
	if (ap == NULL) {
		clientsenderr(sp, cl, rp, "acl %s unknown", rp->aclname);
		rp->state = RSTATE_DONE;
		return;
	}

	if (ap->compactseq > 0) {
		clientsend(sp, cl, rp, CS_CONTINUE);
		clientwrite(sp, cl, "compaction already in progress for acl %s",
		    rp->aclname);
		clientsendeom(sp, cl);
		rp->state = RSTATE_DONE;
		return;
	}

	syslog(LOG_INFO,
	    "clientcompact: started compacting acl %s (lastseq %d)",
	    ap->name, ap->lastseq);

	if (sp->cf->c_incrseq)
		ap->compactseq = sp->cf->c_lowseq;
	else
		ap->compactseq = sp->cf->c_highseq;
	ap->compactclient = cl;

	/*
	 * We want to unlink this request but we can't do it here.
	 * So allocate a new one, copy stuff over and mark this one done.
	 */
	rp2 = new(1, sizeof(*rp), "clientcompact: req");
	*rp2 = *rp;
	rp->aclname = NULL;
	rp->comment = NULL;
	rp->av = NULL;
	rp->state = RSTATE_DONE;

	rp2->state = RSTATE_PENDING;
	rp2->next = NULL;
	ap->compactreq = rp2;
	timerset(&sp->t_compact, 0);
}

char const *helpstr[] = {
	"reload cookie",
	"drop cookie addr [-]",
#ifdef notdef
	"drop cookie net/width [-]",
	"drop cookie net netmask [-]",
#endif
	"dropudpport cookie port acl [-]",
	"droptcpport cookie port acl [-]",
	"restore cookie addr [-]",
#ifdef notdef
	"restore cookie net/width [-]",
	"restore cookie net netmask [-]",
#endif
	"restoreudpport cookie port acl [-]",
	"restoretcpport cookie port acl [-]",
	"state cookie",
	"listacl cookie acl",
	"query cookie addr",
	"sync cookie",
	"compact cookie acl",
	"exit cookie",
	"version cookie",
	"help cookie",
	NULL
};

void
clienthelp(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register const char **pp;

	clientsend(sp, cl, rp, CS_CONTINUE);
	for (pp = helpstr; *pp != NULL; ++pp)
		clientwrite(sp, cl, "# %s", *pp);
	clientsendeom(sp, cl);
	rp->state = RSTATE_DONE;
}

void
clientlistacl(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register int i;
	register struct acl *al;
	register struct acllist *ap;

	/* We only know about configured ACLs */
	ap = aclfindlistbyname(sp->cf, rp->aclname);
	if (ap == NULL) {
		clientsenderr(sp, cl, rp, "acl %s unknown", rp->aclname);
		rp->state = RSTATE_DONE;
		return;
	}

	clientsend(sp, cl, rp, CS_CONTINUE);
	for (i = 0, al = ap->acl; i < ap->acllen; ++i, ++al)
		clientwrite(sp, cl, aclformat(al));
	clientsendeom(sp, cl);
	rp->state = RSTATE_DONE;
}

void
clientprocess(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register int i;

	switch (rp->type) {

	case REQ_UNKNOWN:
		clientsenderr(sp, cl, rp, "unknown command \"%s\"", rp->cmd);
		rp->state = RSTATE_DONE;
		break;

	case REQ_RELOAD:
		++reload;
		clientsend(sp, cl, rp, 0);
		rp->state = RSTATE_DONE;
		break;

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_DROPUDPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
		childdroprestore(sp, cl, rp);
		break;

	case REQ_STATE:
		clientstate(sp, cl, rp);
		break;

	case REQ_LISTACL:
		clientlistacl(sp, cl, rp);
		break;

	case REQ_QUERY:
		clientquery(sp, cl, rp);
		break;

	case REQ_COMPACT:
		clientcompact(sp, cl, rp);
		break;

	case REQ_EXIT:
		clientsend(sp, cl, rp, 0);
		log(LOG_DEBUG, "client #%d exit", cl->n);
		/* Flag that this client is finished */
		++cl->close;
		rp->state = RSTATE_DONE;
		break;

	case REQ_TEST:
		/* Generate lots of output for testing purposes */
		clientsend(sp, cl, rp, CS_CONTINUE);
		for (i = 1; i <= rp->cookie; ++i)
			clientwrite(sp, cl, "%032d", i);
		clientsendeom(sp, cl);
		rp->state = RSTATE_DONE;
		break;

	case REQ_HELP:
		clienthelp(sp, cl, rp);
		break;

	default:
		log(LOG_ERR, "clientprocess: unhandled request type %d",
		    rp->type);
		exit(EX_SOFTWARE);
	}
}

void
clientquery(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register struct acllist *ap;
	register struct acl *al;

	if (rp->acl.type != ATYPE_BLOCKHOST) {
		clientsenderr(sp, cl, rp,
		    "don't know how to handle acl request type %d",
		    rp->acl.type);
		rp->state = RSTATE_DONE;
		return;
	}

	ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
	if (ap == NULL) {
		clientsenderr(sp, cl, rp,
		    "can't find acllist for %s (is the default acl missing?)",
		    addr2str(rp->acl.addr1.addr));
		rp->state = RSTATE_DONE;
		return;
	}

	al = aclfindacl(ap, &rp->acl);
	if (al == NULL) {
		clientsenderr(sp, cl, rp,
		    "%s not found in acl %s",
		    addr2str(rp->acl.addr1.addr), ap->name);
		rp->state = RSTATE_DONE;
		return;
	}

	clientsend(sp, cl, rp, CS_CONTINUE);
	clientwrite(sp, cl, "acl %s", ap->name);
	clientwrite(sp, cl, aclformat(al));
	clientsendeom(sp, cl);
	rp->state = RSTATE_DONE;
}

void
clientsend(register struct state *sp, register struct client *cl,
    register struct req *rp, register int flags)
{
	u_int32_t cookie;
	const char *cmd;
	register struct timeval *tvp;
	struct timeval tv;
	char buf[1024];

	if (rp != NULL) {
		tvp = &rp->cts;
		if (tvp->tv_sec == 0 && tvp->tv_sec == 0)
			getts(tvp);
		cmd = rp->cmd;
		cookie = rp->cookie;
	} else {
		tvp = &tv;
		getts(tvp);
		cmd = "acld";
		cookie = 0;
	}
	(void)snprintf(buf, sizeof(buf), "%ld.%06ld %u %s%s%s",
	    tvp->tv_sec,
	    tvp->tv_usec,
	    cookie,
	    cmd,
	    (flags & CS_FAILED) != 0 ? "-failed" : "",
	    (flags & CS_CONTINUE) != 0 ? " -" : "");
	clientwrite(sp, cl, buf);
}

void
clientsendeom(register struct state *sp, register struct client *cl)
{

	clientwriteraw(sp, cl, ".\r\n");
}

void
clientsenderr(register struct state *sp, register struct client *cl,
    register struct req *rp, register const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	clientsend(sp, cl, rp, CS_CONTINUE | CS_FAILED);
	clientwritevar(sp, cl, fmt, ap);
	clientsendeom(sp, cl);
	va_end(ap);
}

void
clientstate(register struct state *sp, register struct client *cl,
    register struct req *rp)
{
	register struct acllist *ap;
	register struct intlist *ip;
	register struct addr *a;
	register struct cf *cf;
	register int i, j, w;
	register char *cp;
	register const char *p;
	char buf[32];

	cf = sp->cf;
	clientsend(sp, cl, rp, CS_CONTINUE);
	p = val2str(astate2str, sp->state);
	if (p == NULL) {
		(void)snprintf(buf, sizeof(buf), "#%d", sp->state);
		p = buf;
	}
	clientwrite(sp, cl, "# state %s", p);
	clientwrite(sp, cl, "# version %s", version);
	clientwrite(sp, cl, "# config \"%s\"", configfn);
	if (verbose)
		clientwrite(sp, cl, "# verbose %d", verbose);
	if (debug)
		clientwrite(sp, cl, "# debug %d", debug);
	if (cf != NULL) {
		clientwrite(sp, cl, "router \"%s\"", cf->c_router);
		clientwrite(sp, cl, "expect \"%s\"", cf->c_expect);
		clientwrite(sp, cl, "script \"%s\"", cf->c_script);
		clientwrite(sp, cl, "port %d", (int)cf->c_port);
		if (cf->c_netsfac != 0)
			clientwrite(sp, cl, "netsfac %s",
			    val2str(syslog2str, cf->c_netsfac));
		if (cf->c_login_secs != LOGIN_SECS)
			clientwrite(sp, cl, "login_secs %d", cf->c_login_secs);
		if (cf->c_ayt_secs != AYT_SECS)
			clientwrite(sp, cl, "ayt_secs %d", cf->c_ayt_secs);
		if (cf->c_sync_secs != SYNC_SECS)
			clientwrite(sp, cl, "sync_secs %d", cf->c_sync_secs);
		if (cf->c_select_secs != SELECT_SECS)
			clientwrite(sp, cl, "select_secs %d",
			    cf->c_select_secs);
		if (cf->c_lowseq != 0)
			clientwrite(sp, cl, "lowseq %d", cf->c_lowseq);
		if (cf->c_highseq != 0)
			clientwrite(sp, cl, "highseq %d", cf->c_highseq);
		if (cf->c_maxseq != 0)
			clientwrite(sp, cl, "maxseq %d", cf->c_maxseq);
		if (cf->c_lowportseq != 0)
			clientwrite(sp, cl, "lowportseq %d", cf->c_lowportseq);
		if (cf->c_highportseq != 0)
			clientwrite(sp, cl, "highportseq %d",
			    cf->c_highportseq);
		clientwrite(sp, cl, "incrseq %d", cf->c_incrseq);
		for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen;
		    ++i, ++ap)
			clientwrite(sp, cl, "# lastseq for ACL %s is %d",
			    ap->name, ap->lastseq);
		for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen;
		    ++i, ++ap)
			clientwrite(sp, cl, "# acllen for ACL %s is %d",
			    ap->name, ap->acllen);
		for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen;
		    ++i, ++ap)
			if (ap->compactseq > 0)
				clientwrite(sp, cl,
				    "# compactseq for ACL %s is %d",
				    ap->name, ap->compactseq);

		for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen;
		    ++i, ++ap) {
			for (j = 0, a = ap->addr; j < ap->addrlen; ++j, ++a) {
				w = maskwidth(a->mask);
				if (a->addr == 0 && w == 32)
					cp = "default";
				else
					cp = addr2str(a->addr);
				if (w == 32)
					buf[0] = '\0';
				else
					(void)snprintf(buf, sizeof(buf),
					    "/%d", w);
				clientwrite(sp, cl, "acl %s %s%s",
				    ap->name, cp, buf);
			}
		}
		for (i = 0, ip = cf->c_intlist; i < cf->c_intlistlen;
		    ++i, ++ip)
			clientwrite(sp, cl, "interface %s %s",
			ip->i_name, ip->i_acl);
	}
	clientsendeom(sp, cl);
	rp->state = RSTATE_DONE;
}

/* Possibly double up leading dot and adding trailing return/newline */
void
clientwrite(register struct state *sp, register struct client *cl,
    register const char *fmt, ...)
{
	register char *cp, *bp;
	char buf[1 + 8192 + 2];
	va_list ap;

	bp = buf + 1;
	va_start(ap, fmt);
	(void)vsnprintf(bp, sizeof(buf) - (1 + 2), fmt, ap);
	va_end(ap);
	if (*bp == '.')
		*--bp = '.';
	cp = bp + strlen(bp);
	*cp++ = '\r';
	*cp++ = '\n';
	*cp++ = '\0';
	clientwriteraw(sp, cl, bp);
}

/* Queue up output */
void
clientwriteraw(register struct state *sp, register struct client *cl,
    register const char *buf)
{
	register size_t n;
	register ssize_t size;
	register struct iobuf *ip;

	ip = &cl->wbuf;
	n = strlen(buf);
	size = ip->len + n + 1;
	dynarray((void **)&ip->buf, &ip->len, &ip->size,
	    sizeof(*ip->buf), 8192, size, "clientwriteraw iobuf");
	memmove(ip->buf + ip->len, buf, n);
	if (ip->len == 0)
		sp->nfds = 0;
	ip->len += n;
}

/* Possibly double up leading dot and adding trailing return/newline */
void
clientwritevar(register struct state *sp, register struct client *cl,
    register const char *fmt, va_list ap)
{
	register char *cp, *bp;
	char buf[1 + 8192 + 2];

	bp = buf + 1;
	(void)vsnprintf(bp, sizeof(buf) - (1 + 2), fmt, ap);
	if (*bp == '.')
		*--bp = '.';
	cp = bp + strlen(bp);
	*cp++ = '\r';
	*cp++ = '\n';
	*cp++ = '\0';
	clientwriteraw(sp, cl, bp);
}

void
freeclient(register struct state *sp, register struct client *cl)
{
	register struct client *cl2;
	register struct req *rp, *rp2;
	register struct acllist *ap;
	register int i;

	if (cl->c >= 0) {
		close(cl->c);
		cl->c = -1;
	}
	sp->nfds = 0;
	iofree(&cl->rbuf);
	iofree(&cl->wbuf);

	rp = cl->req;
	while (rp != NULL) {
		rp2 = rp->next;
		rp->next = NULL;
		freereq(rp);
		rp = rp2;
	}
	if (cl->treq != NULL)
		freereq(cl->treq);

	if (sp->clients == cl) {
		sp->clients = sp->clients->next;
	} else {
		for (cl2 = sp->clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == cl)
				break;
		if (cl2 == NULL) {
			log(LOG_ERR, "freeclient: client #%d: can't happen",
			    cl->n);
			exit(EX_SOFTWARE);
		}
		cl2->next = cl->next;
	}
	cl->next = NULL;

	if (cl == sp->reqclient)
		sp->reqclient = NULL;
	free(cl);

	/* Clean up compaction state */
	for (i = 0, ap = sp->cf->c_acllist; i < sp->cf->c_acllistlen; ++i, ++ap)
		if (ap->compactclient == cl) {
			ap->compactclient = NULL;
			if (ap->compactreq != NULL) {
				freereq(ap->compactreq);
				ap->compactreq = NULL;
			}
			break;
		}
}

struct client *
newclient(register struct state *sp, register int c)
{
	register struct client *cl, *cl2;

	cl = (struct client *)new(1, sizeof(*cl), "newclient cl");
	cl->c = c;
	cl->n = sp->nclients++;
	sp->nfds = 0;

	/* Append to end of clinet list */
	if (sp->clients == NULL) {
		sp->clients = cl;
	} else {
		for (cl2 = sp->clients; cl2 != NULL; cl2 = cl2->next)
			if (cl2->next == NULL)
				break;
		cl2->next = cl;
	}
	return (cl);
}
