#!/bin/sh
# @(#) $Id: createhostcert.sh,v 1.6 2004/03/25 00:53:07 leres Exp $ (LBL)
#
# createhost.sh - create a host key and certificate
#
# usage:
#
usage="usage: ${prog} [-f] fqdn rootca.key rootca.pem serial host.key host.pem"
#
#	-f - overwrite existing key and cert
#
# This script creates a host key and certificate.
# It will refuse to overwrite an existing key unless -f is specified.
#

PATH="/usr/ssl/bin:/usr/local/ssl/bin:${PATH}"
export PATH

prog="`basename $0`"

t1=/tmp/${prog}.1.$$
t2=/tmp/${prog}.2.$$
t3=/tmp/${prog}.3.$$

trap 'rm -f ${t1} ${t2} ${t3}; exit 1' 1 2 3 15

PRIVATEUMASK="umask 077"

force=0

set -- `getopt f $*`

for a do
	case "${a}" in

	-f)
		force=1
		shift
		;;

	--)
		shift
		break
		;;
	esac
done

if [ $# -ne 6 ]; then
	echo "${usage}" 1>&2
	exit 1
fi


fqdn=$1
rootkey=$2
rootpem=$3
serial=$4
key=$5
pem=$6

privatedir="`dirname ${key}`"
certdir="`dirname ${pem}`"

database="${certdir}/rootca.index"

if [ ${force} -eq 0 ]; then
	if [ -f ${key} ]; then
		echo "${prog}: ${key} already exists (aborting)" 1>&2
		exit 1
	fi
	if [ -f ${pem} ]; then
		echo "${prog}: ${pem} already exists (aborting)" 1>&2
		exit 1
	fi
fi

if [ "${privatedir}" != "." -a ! -d ${privatedir} ]; then
	(${PRIVATEUMASK} ; mkdir ${privatedir})
	if [ $? -ne 0 ]; then
		echo "${prog}: (aborting)" 1>&2
		exit 1
	fi
fi

if [ ${force} -ne 0 ]; then
	rm -f ${key} ${pem}
fi

echo ">> Create a key for ${fqdn}"
(${PRIVATEUMASK} ; set -x ; openssl genrsa -out ${key} 1024)
if [ $? -ne 0 ]; then
	rm -f ${t1} ${t2} ${t3}
	exit 1
fi

if [ "${certdir}" != "." -a ! -d ${certdir} ]; then
	mkdir ${certdir}
	if [ $? -ne 0 ]; then
		echo "${prog}: (aborting)" 1>&2
		rm -f ${t1} ${t2} ${t3}
		exit 1
	fi
fi

cat <<EOF >${t1}
[ req ]
prompt				= no
default_bits			= 1024
distinguished_name		= req_distinguished_name
string_mask			= nombstr
req_extensions			= v3_req

[ req_distinguished_name ]
countryName			= ZZ
stateOrProvinceName		= Atlantis
localityName			= The Village
0.organizationName		= Unsuitable for security at any keysize
organizationalUnitName		= Bogon Anonymous Certificate Authority
commonName			= ${fqdn}
emailAddress			= root@${fqdn}

[ v3_req ]
nsCertType			= client
basicConstraints		= critical,CA:false
EOF


echo ">> Create a certificate request for ${fqdn}"
(set -x ;  openssl req -new -config ${t1} -key ${key} -out ${t2})
if [ $? -ne 0 ]; then
	rm -f ${t1} ${t2} ${t3}
	exit 1
fi

if [ ! -f ${database} ]; then
	touch ${database}
fi

oldserialnum="`cat ${serial}`"

cat <<EOF >${t3}
[ ca ]
default_ca = default_CA

[ default_CA ]
dir = ${certdir}
certs = ${certdir}
new_certs_dir = ${certdir}
database = ${database}
serial = ${serial}
certificate = ${rootpem}
private_key = ${rootkey}
default_days = 3650
default_md = md5
preserve = yes
x509_extensions = user_cert
policy = policy_anything

[ policy_anything ]
commonName = supplied
emailAddress = supplied

[ user_cert ]
authorityKeyIdentifier = keyid:always
EOF

echo ">> Sign the certificate request for ${fqdn} using the rootca"
(set -x ; (echo yes ; echo yes) |
    openssl ca -config ${t3} -out ${pem} -infiles ${t2})

if [ $? -ne 0 ]; then
	rm -f ${t1} ${t2} ${t3}
	exit 1
fi

echo ">> Verify the the certificate request for ${fqdn} using the rootca"
(set -x ; openssl verify -CAfile ${rootpem} ${pem})
if [ $? -ne 0 ]; then
	rm -f ${t1} ${t2} ${t3}
	exit 1
fi

echo ">> Removing junk we don't need"
(set -x ; rm -f ${certdir}/${oldserialnum}.pem ${serial}.old \
    ${database}.old ${certdir}/rootca.index.attr)

echo ">> Rehashing the cert directory (${certdir})"
(set -x ; c_rehash ${certdir})
if [ $? -ne 0 ]; then
	rm -f ${t1} ${t2} ${t3}
	exit 1
fi

echo ">> Successful!"

rm -f ${t1} ${t2} ${t3}
