/*
 * Copyright (c) 2004
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: server.c,v 1.8 2004/08/12 21:10:20 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "cf.h"
#include "child.h"
#include "server.h"
#include "util.h"

/* Forwards */
void servercompactone(struct state *, struct acllist *);
void serverdroprestore(struct state *, struct req *);
void servermoveacl(struct state *, struct acl *, int);

void
serverayt(register struct state *sp)
{
	register struct req *rp;

	rp = new(1, sizeof(*rp), "serverayt: req");
	rp->state = RSTATE_PENDING;
	rp->type = REQ_AYT;
	rp->cmd = "ayt";

	/* Arrival timestamp */
	getts(&rp->ats);

	/* Append the new request to end of list */
	appendreq(&sp->req, rp);
}

void
servercompact(register struct state *sp)
{
	register struct acllist *ap;
	register struct cf *cf;
	register int i;

	cf = sp->cf;
	for (i = 0, ap = cf->c_acllist; i < cf->c_acllistlen; ++i, ++ap)
		if (ap->compactseq > 0) {
			servercompactone(sp, ap);
			break;
		}
}

void
servercompactone(register struct state *sp, register struct acllist *ap)
{
	register int n, inc, seq;
	register struct acl *al;

	n = ap->acllen;
	if (sp->cf->c_incrseq) {
		al = ap->acl;
		inc = 1;
	} else {
		al = ap->acl + n - 1;
		inc = -1;
	}

	for (; n > 0; --n, al += inc) {
		if (al->seq < sp->cf->c_lowseq || al->seq > sp->cf->c_highseq)
			continue;
		if (sp->cf->c_incrseq) {
			if (al->seq < ap->compactseq)
				continue;
		} else {
			if (al->seq > ap->compactseq)
				continue;
		}
		if (al->type == ATYPE_UNKNOWN) {
			ap->compactseq = 0;
			timerreset(&sp->t_compact);
			log(LOG_ERR, "servercompactone: Can't handle"
			    " unknown acl type (seq %d)", al->seq);
			return;
		}
		seq = ap->compactseq;
		ap->compactseq += inc;
		if (al->seq != seq) {
			/* XXX Check to see if it's in use? */
			servermoveacl(sp, al, seq);

			/* We're not done yet so reschedule */
			timerset(&sp->t_compact, 0);
			return;
		}

		/* XXX paranoid */
		if (sp->cf->c_incrseq) {
			if (ap->compactseq >= sp->cf->c_highseq)
				abort();
		} else {
			if (ap->compactseq <= sp->cf->c_lowseq)
				abort();
		}
	}

	/* If we got here we're done */
	ap->lastseq = ap->compactseq - inc;
	ap->compactseq = 0;
	timerreset(&sp->t_compact);

	/* Schedule a sync */
	if (sp->cf->c_sync_secs > 0 && timercheck(&sp->t_sync) < 0)
		timerset(&sp->t_sync, sp->cf->c_sync_secs);

	log(LOG_INFO, "servercompact: finished compacting acl %s (lastseq %d)",
	    ap->name, ap->lastseq);

	/* Tell client the good news */
	if (ap->compactclient != NULL) {
		if (ap->compactreq == NULL)
			abort();
		clientsend(sp, ap->compactclient, ap->compactreq, 0);
		freereq(ap->compactreq);
		ap->compactclient = NULL;
		ap->compactreq = NULL;
	}
}

void
servermoveacl(register struct state *sp, register struct acl *al,
    register int seq)
{
	register struct req *rp;
	int rtype, dtype;

	switch (al->type) {

	case ATYPE_BLOCKHOST:
		rtype = REQ_RESTORE;
		dtype = REQ_DROP;
		break;

	case ATYPE_BLOCKUDPPORT:
		rtype = REQ_RESTOREUDPPORT;
		dtype = REQ_DROPUDPPORT;
		break;

	case ATYPE_BLOCKTCPPORT:
		rtype = REQ_RESTORETCPPORT;
		dtype = REQ_DROPTCPPORT;
		break;

	default:
		log(LOG_ERR, "servermoveacl: bad type %d (1)", al->type);
		exit(EX_SOFTWARE);
	}

	rp = new(1, sizeof(*rp), "servermoveacl: req 1");
	rp->state = RSTATE_PENDING;
	rp->type = rtype;
	rp->cmd = val2str(cmd2req, rp->type);
	rp->acl = *al;
	rp->comment = strsave("compact");
	getts(&rp->ats);
	appendreq(&sp->req, rp);

	/* Change the sequence number */
	al->seq = seq;

	rp = new(1, sizeof(*rp), "servermoveacl: req 2");
	rp->state = RSTATE_PENDING;
	rp->type = dtype;
	rp->cmd = val2str(cmd2req, rp->type);
	rp->acl = *al;
	rp->comment = strsave("compact");
	getts(&rp->ats);
	appendreq(&sp->req, rp);
}

/* XXX unfortunately, this duplicates a lot of childdroprestore() */
void
serverdroprestore(register struct state *sp, register struct req *rp)
{
	register struct acllist *ap;
	register struct acl *al;

	/* XXX server should normally supply the sequence number */
	if (rp->acl.seq == 0) {
		log(LOG_ERR, "serverdroprestore: seq number shouldn't be zero");
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
		return;
	}

	/* Find acl list */
	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
		ap = aclfindlistbyaddr(sp->cf, &rp->acl.addr1);
		if (ap == NULL) {
			log(LOG_ERR,
			    "serverdroprestore: can't find acllist for %s",
			    addr2str(rp->acl.addr1.addr));
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		break;

	case ATYPE_BLOCKUDPPORT:
	case ATYPE_BLOCKTCPPORT:
		ap = aclfindlistbyname(sp->cf, rp->aclname);
		if (ap == NULL) {
			log(LOG_ERR,
			    "serverdroprestore: can't find acllist for %s",
			    rp->acl.port1);
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		break;

	default:
		log(LOG_ERR, "Acl type %d not implemented", rp->acl.type);
		rp->state = RSTATE_DONE;
		getts(&rp->cts);
		return;
	}

	/* Check if acl should or should not already exist */
	al = aclfindacl(ap, &rp->acl);
	switch (rp->type) {

	case REQ_DROP:
		if (al == NULL) {
			log(LOG_ERR,
			    "%s wasn't is already blocked in acl %s!",
			    addr2str(rp->acl.addr1.addr), ap->name);
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
		if (al == NULL) {
			log(LOG_ERR, "serverdroprestore:"
			    " %s wasn't already blocked in acl %s!",
			    rp->acl.port1, ap->name);
			rp->state = RSTATE_DONE;
			getts(&rp->cts);
			return;
		}
		break;

	case REQ_RESTORE:
		if (al == NULL) {
			log(LOG_ERR,
			    "serverdroprestore: wasn't %s found in acl %s",
			    addr2str(rp->acl.addr1.addr), ap->name);
			getts(&rp->cts);
			rp->state = RSTATE_DONE;
			return;
		}
		break;


	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
		if (al == NULL) {
			log(LOG_ERR,
			    "serverdroprestore: wasn't %s not found in acl %s",
			    rp->acl.port1, ap->name);
			getts(&rp->cts);
			rp->state = RSTATE_DONE;
			return;
		}
		break;

	default:
		log(LOG_ERR, "serverdroprestore: bad type %d (1)", rp->type);
		exit(EX_SOFTWARE);
	}

	/* No need to check the sequence number further, just check the type */
	switch (rp->type) {

	case REQ_DROP:
	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORE:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
		break;

	default:
		log(LOG_ERR, "serverdroprestore: bad type %d (2)", rp->type);
		exit(EX_SOFTWARE);
	}

	/* Send the request to the child */
	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
		childsend(sp, "%s %s %s %d",
		    rp->cmd, addr2str(rp->acl.addr1.addr),
		    ap->name, rp->acl.seq);
		break;

	case ATYPE_BLOCKUDPPORT:
	case ATYPE_BLOCKTCPPORT:
		childsend(sp, "%s %s %s %d",
		    rp->cmd, rp->acl.port1, ap->name, rp->acl.seq);
		break;
			
	default:
		log(LOG_ERR, "childdroprestore: bad atype %d", rp->acl.type);
		exit(EX_SOFTWARE);
	}

	/* Update request state */
	rp->state = RSTATE_CHILD;
}

void
serverprocess(register struct state *sp, register struct req *rp)
{

	switch (rp->type) {

	case REQ_DROP:
	case REQ_RESTORE:
	case REQ_DROPUDPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_RESTORETCPPORT:
		serverdroprestore(sp, rp);
		break;

	case REQ_AYT:
	case REQ_SYNC:
		childsend(sp, "%s", rp->cmd);
		rp->state = RSTATE_CHILD;
		break;

	default:
		log(LOG_ERR, "serverprocess: unhandled request type %d",
		    rp->type);
		exit(EX_SOFTWARE);
	}
}

void
serversync(register struct state *sp)
{
	register struct req *rp;

	rp = new(1, sizeof(*rp), "serversync: req");
	rp->state = RSTATE_PENDING;
	rp->type = REQ_SYNC;
	rp->cmd = "sync";

	/* Arrival timestamp */
	getts(&rp->ats);

	/* Append the new request to end of list */
	appendreq(&sp->req, rp);

	/* Cancel possible pending "ayt" */
	timerreset(&sp->t_ayt);
}
