/*
 * Copyright (c) 2002, 2003, 2004
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: util.c,v 1.7 2004/09/13 17:49:55 leres Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>

#include "acld.h"
#include "util.h"

struct s2v syslog2str[] = {
        { "auth",       LOG_AUTH, },
        { "authpriv",   LOG_AUTHPRIV, },
        { "console",    LOG_CONSOLE, },
        { "cron",       LOG_CRON, },
        { "daemon",     LOG_DAEMON, },
        { "ftp",        LOG_FTP, },
        { "kern",       LOG_KERN, },
        { "lpr",        LOG_LPR, },
        { "mail",       LOG_MAIL, },
        { "news",       LOG_NEWS, },
        { "ntp",        LOG_NTP, },
        { "security",   LOG_SECURITY, },
        { "syslog",     LOG_SYSLOG, },
        { "user",       LOG_USER, },
        { "uucp",       LOG_UUCP, },
        { "local0",     LOG_LOCAL0, },
        { "local1",     LOG_LOCAL1, },
        { "local2",     LOG_LOCAL2, },
        { "local3",     LOG_LOCAL3, },
        { "local4",     LOG_LOCAL4, },
        { "local5",     LOG_LOCAL5, },
        { "local6",     LOG_LOCAL6, },
        { "local7",     LOG_LOCAL7, },
        { NULL,         -1, }
};

char *
addr2str(register u_int32_t a)
{
	struct in_addr in;

	in.s_addr = htonl(a);
	return (inet_ntoa(in));
}

void
dynarray(register void **p, register int *lenp, register int *sizep,
    register size_t osize, register int isize, register int inc,
    register const char *what)
{

	if (*p == NULL) {
		if (*sizep <= 0)
			*sizep = isize;
		*p = new(*sizep, osize, what);
	}
	if (*sizep <= (*lenp + inc/2)) {
		*sizep += inc;
		*p = realloc(*p, *sizep * osize);
		if (*p == NULL) {
			log(LOG_ERR, "dynarray: realloc %s: %s",
			    what, strerror(errno));
			exit(EX_OSERR);
		}
		memset(*p + *lenp * osize, 0, (*sizep - *lenp) * osize);
	}
}

const char *
escapestr(const char *str)
{
	register char *cp;
	register const char *p;
	register size_t n;
	static char *result = NULL;
	static char escapes[] = "$[]{}\"\\";	/* XXX check this in book */

	if (result != NULL) {
		free(result);
		result = NULL;
	}

	/* Return a expect style null */
	if (str == NULL || *str == '\0')
		return ("{}");

	/* Enough space for string and EOS */
	n = strlen(str) + 1;

	/* Add room for escapes */
	for (p = str; *p != '\0'; ++p)
		if (strchr(escapes, *p) != NULL)
			++n;

	result = malloc(n);
	if (result == NULL) {
		fprintf(stderr, "%s: escapenaem: malloc: %s",
		    prog, strerror(errno));
		exit(EX_OSERR);
	}
	p = str;
	cp = result;
	while (*p != '\0') {
		if (strchr(escapes, *p) != NULL)
			*cp++ = '\\';
		*cp++ = *p++;
	}
	*cp = '\0';

	return (result);
}

const char *
extractaddr(register const char *s1, register const char *s2,
    register struct addr *addrp)
{
	register char *cp;
	register const char *p;
	char temp[132];
	int masklen;
	struct in_addr in;

	addrp->addr = 0;
	addrp->mask = 0;

	if (strcmp(s1, "default") == 0) {
		if (s2 != NULL)
			return("default with s2 not null");
		addrp->mask = 0xffffffff;
		return (NULL);
	}

	p = strchr(s1, '/');
	if (p != NULL) {
		/* Mask length specified */
		if (s2 != NULL)
			return ("garbage following masklen");
		masklen = atoi(p + 1);
		if (masklen > 32)
			return ("mask length must be <= 32");
		addrp->mask = 0xffffffff << (32 - masklen);

		/* Get addr */
		strncpy(temp, s1, sizeof(temp));
		temp[sizeof(temp) - 1] = '\0';
		cp = strchr(temp, '/');
		if (cp != NULL)
			*cp = '\0';
		if (!inet_aton(temp, &in))
			return ("bad address");
		addrp->addr = ntohl(in.s_addr);
	} else {
		/* Address and optional mask */
		if (!inet_aton(s1, &in))
			return ("bad address");
		addrp->addr = ntohl(in.s_addr);
		if (s2 == NULL)
			addrp->mask = 0xffffffff;
		else {
			if (!inet_aton(s2, &in))
				return ("bad mask");
			addrp->mask = ntohl(in.s_addr);
		}
	}

	if ((addrp->addr & ~addrp->mask) != 0)
		return ("non-network bits set in addr");

	if ((addrp->addr & 0xff000000) == 0)
		return ("high octet must be non-zero");

	return (NULL);
}

void
freeargv(register char **av)
{
	register char **av2;

	if (av != NULL) {
		av2 = av;
		while (*av2 != NULL)
			free(*av2++);
		free(av);
	}
}

void
freedynarray(register void **p, register int *lenp, register int *sizep)
{
	if (*p != NULL) {
		free(*p);
		*p = NULL;
	}
	*lenp = 0;
	*sizep = 0;
}

void
getts(register struct timeval *tsp)
{

	if (gettimeofday(tsp, NULL) < 0) {
		log(LOG_ERR, "getts: gettimeofday: %s", strerror(errno));
		exit(EX_OSERR);
	}
}

void
log(register int pri, const char *fmt, ...)
{
	va_list ap;

	if (debug) {
		fputs(tsstr(), stderr);
		putc(' ', stderr);
		va_start(ap, fmt);
		(void)vfprintf(stderr, fmt, ap);
		putc('\n', stderr);
		va_end(ap);
	}

	va_start(ap, fmt);
	vsyslog(pri, fmt, ap);
	va_end(ap);
}

int
maskwidth(register u_int32_t m)
{
	register int w;
	register u_int32_t tm;

	for (w = 32; w > 0; --w) {
		tm = 0xffffffff << (32 - w);
		if (tm == m)
			break;
	}
	return (w);

}

void *
new(register size_t num, register size_t size, register const char *what)
{
	register void *p;

	p = calloc(num, size);
	if (p == NULL) {
		log(LOG_ERR, "new: calloc %s: %s", what, strerror(errno));
		exit(EX_OSERR);
	}
	return (p);
}

const char *
pretty(register const char *str, register ssize_t len)
{
	register char *cp;
	register const char *p;
	int i, left;
	static char buf[8192];

	cp = buf;
	left = sizeof(buf) - 2;
	for (i = 0, p = str; i < len; ++i, ++p) {
		if (left < 3)
			break;
		switch (*p) {

		case '\n':
			*cp++ = '\\';
			*cp++ = 'n';
			left -= 2;
			break;

		case '\r':
			*cp++ = '\\';
			*cp++ = 'r';
			left -= 2;
			break;

		default:
			if (isprint((int)*p)) {
				*cp++ = *p;
				--left;
			} else {
				/* DEL to ?, others to alpha */
				*cp++ = '^';
				*cp++ = *p ^ 0x40;
				left -= 2;
			}
			break;
		}
	}
	*cp = '\0';
	return (buf);
}

const char *
simplewd(void)
{
	register char *cp;
	static char wd[1024];

	if (getcwd(wd, sizeof(wd)) == NULL) {
		log(LOG_ERR, "simplewd: getcwd: %s", strerror(errno));
		exit(EX_OSERR);
	}
	cp = strrchr(wd, '/');
	if (cp == NULL)
		return (wd);
	return (++cp);
}

int
suck2dot(register FILE *f, char **pp)
{
	register int n, size, len, i;
	char buf[1024];
	register char *p, *cp;

	n = 0;
	size = 256;
	len = 0;
	p = (char *)malloc(size);
	if (p == NULL) {
		fprintf(stderr, "%s: suck2dot: malloc: %s",
		    prog, strerror(errno));
		exit(EX_OSERR);
	}
	*p = '\0';
	*pp = p;
	while (fgets(buf, sizeof(buf), f) != NULL) {
		++n;
		cp = buf;
		/* Eat leading dot; done if there was only one */
		if (*cp == '.' && *++cp == '\n' && cp[1] == '\0')
			break;

		i = strlen(cp);
		if (len + i + 1 > size) {
			size += len + 256;
			p -= (u_int)*pp;
			*pp = (char *)realloc(*pp, size);
			if (*pp == NULL) {
				log(LOG_ERR, "suck2dot: realloc: %s",
				    strerror(errno));
				exit(EX_OSERR);
			}
			p += (u_int)*pp;
		}
		strcpy(p, cp);
		p += i;
		len += i;
	}
	return (n);
}

int
str2val(register const struct s2v *lp, register const char *suffix)
{
	register const struct s2v *tp;

	for (tp = lp; tp->s != NULL; ++tp)
		if (strcmp(suffix, tp->s) == 0)
			break;

	return (tp->v);
}

int
str2valcase(register const struct s2v *lp, register const char *suffix)
{
	register const struct s2v *tp;

	for (tp = lp; tp->s != NULL; ++tp)
		if (strcasecmp(suffix, tp->s) == 0)
			break;

	return (tp->v);
}

void
trimws(register char *str)
{
	register char *cp;

	cp = str + strlen(str);
	while (cp > str && isspace((int)cp[-1]))
		--cp;
	*cp = '\0';
}

const char *
tsstr()
{
	struct timeval tv;
	static char buf[32];

	getts(&tv);
	snprintf(buf, sizeof(buf), "%ld.%06ld", tv.tv_sec, tv.tv_usec);
	return (buf);
}

int
typ2flag(register const struct v2v *lp, register int wtype)
{
	register const struct v2v *vp;

	for (vp = lp; vp->v1 != 0; ++vp)
		if (vp->v1 == wtype)
			break;
	return (vp->v2);
}

const char *
val2str(register const struct s2v *lp, register int val)
{
	register const struct s2v *tp;

	for (tp = lp; tp->s != NULL; ++tp)
		if (val == tp->v)
			break;

	return (tp->s);
}


int
makeargv(register const char *args, register char ***avp)
{
	register const char *p, *p2, *ep;
	register int n, len;
	register char **av;

	if (args == NULL)
		return (0);

	/* Setup end pointer (trimming trailing whitespace) */
	p = args;
	ep = args + strlen(args);
	while (ep - 1 >= args && isspace((int)ep[-1]))
		--ep;

	/* Eat leading whitespace */
	while (isspace((int)*args))
		++args;
	/* Count arguments */
	n = 0;
	while (p < ep) {
		if (!isspace((int)*p)) {
			++n;
			while (p < ep && !isspace((int)*p))
				++p;
		}
		while (p < ep && isspace((int)*p))
			++p;
	}

	av = new(n + 1, sizeof(*av), "makeargs av");
	*avp = av;

	/* Copy arguments */
	n = 0;
	p = args;
	p2 = p;
	while (p < ep) {
		if (!isspace((int)*p)) {
			++n;
			while (p < ep && !isspace((int)*p))
				++p;
			len = p - p2;
			*av = new(len + 1, sizeof(char), "makeargs string");
			strncpy(*av, p2, len);
			av[0][len] = '\0';
			++av;
		}
		while (p < ep && isspace((int)*p))
			++p;
		p2 = p;
	}

	return (n);
}

void
strappend(register char **pp, register const char *str)
{
	register int len;

	if (*pp == NULL)
		*pp = strsave(str);
	else {
		len = strlen(*pp);
		*pp = realloc(*pp, len + strlen(str) + 1);
		if (*pp == NULL) {
			log(LOG_ERR, "strappend: realloc: %s", strerror(errno));
			exit(EX_OSERR);
		}
		strcpy(*pp + len, str);
	}
}

char *
strsave(register const char *str)
{
	register char *cp;

	cp = strdup(str);
	if (cp == NULL) {
		log(LOG_ERR, "strsave: strdup: %s", strerror(errno));
		exit(EX_OSERR);
	}
	return (cp);
}
